#include "MSSMparam.h"
#include "kinematics.h"
#include "pdefs.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <wait.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>

/*#define MW	pars[3]
#define MZ	pars[4]
#define ME	pars[7]
#define ML	pars[8]
#define MB	pars[9]
#define MT	pars[10]
#define MA0	pars[13]
#define TB	pars[17]*/

double CW, SW, 
       Mh0, MHH, MHp,
       CA, SA, C2A, S2A, CB, SB, C2B, S2B, CAB, SAB, CBA, SBA,
       MSf111, MSf121, MSf221, MSf131, MSf231, MSf141, MSf241, 
       MSf112, MSf122, MSf222, MSf132, MSf232, MSf142, MSf242,
       MSf113, MSf123, MSf223, MSf133, MSf233, MSf143, MSf243,
       USf2311, USf2312, USf2321, USf2322, 
       USf3311, USf3312, USf3321, USf3322, 
       USf4311, USf4312, USf4321, USf4322, 
       MCha1, MCha2,
       UCha11, UCha12, UCha21, UCha22, VCha11, VCha12, VCha21, VCha22,
       MNeu1, MNeu2, MNeu3, MNeu4,
       ZNeu11, ZNeu12, ZNeu13, ZNeu14, ZNeu21, ZNeu22, ZNeu23, ZNeu24, 
       ZNeu31, ZNeu32, ZNeu33, ZNeu34, ZNeu41, ZNeu42, ZNeu43, ZNeu44;


double calculate_mw(const char *expath)
{
  char fname[L_tmpnam], oname[L_tmpnam];
  char line[100];
  int pid;
  ofstream of(tmpnam(fname), ios::out);
  double mwval = 0.0;
  
  strcpy(line, expath);
  strcpy(rindex(line, '/')+1, "pomssmSPA.exe");
  
  if(! of)
  {
    cerr << "Error when writing into temporary file " << fname << endl;
    return 0;
  }      
  of.precision(8);
  of << "1\n1\n1\n1\n1\n";
  of << sqrt(mus) << endl;
  of << MSq3L << endl << MSu3R << endl << MSd3R << endl 
	  << MSl3L << endl << MSl3R << endl;
  of << At << endl << Ab << endl << Atau << endl;
  of << M1 << endl << M2 << endl << M2 << endl << mu << endl;
	  			//  ^^ M3 !!!!!!!!!!!
  of << MA0 << endl << TB << endl << MT << endl << MB << endl;
  of << "2" << endl;
  of.close();
  strcat(line, " <");
  strcat(line, fname);
  strcat(line, " >");
  strcat(line, tmpnam(oname));
  if((pid = fork()) == 0)
  {
    if(execl("/bin/sh", "/bin/sh", "-c", line, NULL) < 0)
    {
      cout << "Error on execuation of pomssmSPA!" << endl;
      exit(1);
    }
  }
  else
  {
    if(pid < 0)
      cerr << "Error on starting child process!" << endl;
    while(wait(NULL) != pid);
    ifstream in(oname, ios::in);
    if(! in)
    {
      cerr << "Error when reading temporary file " << oname << endl;
      return 0;
    }      
    while(in.getline(line, sizeof(line)-1))
    {
      if(strncmp(line, " MW (POMSSM):  ", 15) == 0)
	sscanf(line, "%*s %*s %lf", &mwval);
    }
    in.close();
  }
  remove(fname);
  remove(oname);
  return mwval;
}


void normalize(unsigned len, double *vec)
{
  double vnorm;
  unsigned i;
  
  for(vnorm = 0.0, i = 0; i < len; i++)
    vnorm += vec[i]*vec[i];
  vnorm = sqrt(vnorm);
  for(i = 0; i < len; i++)
    vec[i] /= vnorm;
}

void eigenvalues(double m11, double m12, double m21, double m22,
		 double& ev1, double& ev2)
{
  double sq = sqrt(sqr(m11-m22) + 4*m12*m21);
  ev1 = (m11+m22 - sq)/2.;
  ev2 = (m11+m22 + sq)/2.;
}
       
#define ROTATE(a,i,j,k,l) { g=a[i][j];h=a[k][l];a[i][j]=g-s*(h+g*tau); \
				a[k][l]=h+s*(g-h*tau); }

void setsusypars(const char *refpath)
{
  double GF, MZs, MA0s, SWs, MCha1s, MCha2s,
    	 MCs11, MCs12, MCs21, MCs22,
    	 m11, m12, m22, m1s, m2s;
  double U1[2], U2[2];
/*  double MSl1L = 196.64, MSl1R = 136.23,
	 MSl2L = 196.64, MSl2R = 136.23,
	 MSl3L = 195.75, MSl3R = 133.55,
	 MSq1L = 539.86, MSu1R = 521.66, MSd1R = 519.53,
	 MSq2L = 539.86, MSu2R = 521.66, MSd2R = 519.53,
         MSq3L = 495.91, MSu3R = 424.83, MSd3R = 516.86,
    	 Atau = -254.20, At = -510.01, Ab = -772.66,
    	 M1 = 99.13, M2 = 192.74, mu = 352.39;*/
  
  if(MW < 1.0)
  {
    GF = MW;
    MW = calculate_mw(refpath);
    MW += 115.0 * (MZ/91.187 -1.0);	   // correct for fixed MZ in pomssmSPA
    MW += 1.27865e6 * (GF-0.000011663399); // correct for fixed GF in pomssmSPA
  }
  CW = MW/MZ;
  SW = sqrt(1-CW*CW);
  MA0s = MA0*MA0;
  MZs = MZ*MZ;
  CB = 1/sqrt(1+TB*TB);
  SB = TB*CB;
  C2B = (1-TB*TB)/(1+TB*TB);
  S2B = 2*TB/(1+TB*TB);
  Mh0 = sqrt((MA0s + MZs - sqrt(sqr(MA0s+MZs) - 4*MZs*MA0s*C2B*C2B))/2.);
  MHH = sqrt((MA0s + MZs + sqrt(sqr(MA0s+MZs) - 4*MZs*MA0s*C2B*C2B))/2.);
  MHp = sqrt(MA0s + MW*MW);
  C2A = (MZs-MA0s) / sqrt(sqr(MA0s+MZs) - 4*MZs*MA0s*C2B*C2B) * C2B;
  S2A =(-MZs-MA0s) / sqrt(sqr(MA0s+MZs) - 4*MZs*MA0s*C2B*C2B) * S2B;
  SA = -sqrt((1-C2A)/2.);
  CA = SA * (1+C2A)/S2A;
  CAB = CA*CB-SA*SB;
  SAB = SA*CB+CA*SB;
  CBA = CA*CB+SA*SB;
  SBA = SB*CA-CB*SA;
  SWs = SW*SW;
  MSf111 = sqrt(MSl1L*MSl1L + 1/2.*C2B*MZs);
  MSf121 = sqrt(MSl1L*MSl1L + (-1/2.+SWs)*C2B*MZs);
  MSf221 = sqrt(MSl1R*MSl1R - SWs*C2B*MZs);
  MSf131 = sqrt(MSq1L*MSq1L + (1/2.-2/3.*SWs)*C2B*MZs);
  MSf231 = sqrt(MSu1R*MSu1R + 2/3.*SWs*C2B*MZs);
  MSf141 = sqrt(MSq1L*MSq1L + (-1/2.+1/3.*SWs)*C2B*MZs);
  MSf241 = sqrt(MSd1R*MSd1R - 1/3.*SWs*C2B*MZs);
  MSf112 = sqrt(MSl2L*MSl2L + 1/2.*C2B*MZs);
  MSf122 = sqrt(MSl2L*MSl2L + (-1/2.+SWs)*C2B*MZs);
  MSf222 = sqrt(MSl2R*MSl2R - SWs*C2B*MZs);
  MSf132 = sqrt(MSq2L*MSq2L + (1/2.-2/3.*SWs)*C2B*MZs);
  MSf232 = sqrt(MSu2R*MSu2R + 2/3.*SWs*C2B*MZs);
  MSf142 = sqrt(MSq2L*MSq2L + (-1/2.+1/3.*SWs)*C2B*MZs);
  MSf242 = sqrt(MSd2R*MSd2R - 1/3.*SWs*C2B*MZs);
  MSf113 = sqrt(MSl3L*MSl3L + 1/2.*C2B*MZs);
  
  MCs11 = M2*M2 + 2*MW*MW*SB*SB;
  MCs12 = M_SQRT2*MW*(CB*M2 + mu*SB);
  MCs21 = MCs12;
  MCs22 = mu*mu + 2*MW*MW*CB*CB;
  eigenvalues(MCs11, MCs12, MCs21, MCs22, MCha1s, MCha2s);
  U1[1] = MCs11 - MCha1s;
  U1[0] = -MCs12;
  U2[1] = MCs11 - MCha2s;
  U2[0] = -MCs12;
  normalize(2, U1);
  normalize(2, U2);
  MCha1 = sqrt(MCha1s);
  MCha2 = sqrt(MCha2s);
  UCha11 = U1[0];  UCha12 = U1[1];
  UCha21 = U2[0];  UCha22 = U2[1];
  VCha11 = (M2*UCha11 + M_SQRT2*MW*CB*UCha12) / MCha1;
  VCha21 = (M2*UCha21 + M_SQRT2*MW*CB*UCha22) / MCha2;
  VCha12 = (M_SQRT2*MW*SB*UCha11 + mu*UCha12) / MCha1;
  VCha22 = (M_SQRT2*MW*SB*UCha21 + mu*UCha22) / MCha2;
  
  m11 = MSl3L*MSl3L + ML*ML + (-1/2.+SWs)*C2B*MZs;
  m12 = ML * (Atau - mu*TB);
  m22 = MSl3R*MSl3R + ML*ML - SWs*C2B*MZs;
  eigenvalues(m11, m12, m12, m22, m1s, m2s);
  U1[1] = m11 - m1s;
  U1[0] = -m12;
  U2[1] = m11 - m2s;
  U2[0] = -m12;
  normalize(2, U1);
  normalize(2, U2);
  MSf123 = sqrt(m1s);
  MSf223 = sqrt(m2s);
  USf2311 = U1[0];  USf2312 = U1[1];
  USf2321 = U2[0];  USf2322 = U2[1];
  
  m11 = MSq3L*MSq3L + MT*MT + (1/2.-2/3.*SWs)*C2B*MZs;
  m12 = MT * (At - mu/TB);
  m22 = MSu3R*MSu3R + MT*MT + 2/3.*SWs*C2B*MZs;
  eigenvalues(m11, m12, m12, m22, m1s, m2s);
  U1[1] = m11 - m1s;
  U1[0] = -m12;
  U2[1] = m11 - m2s;
  U2[0] = -m12;
  normalize(2, U1);
  normalize(2, U2);
  MSf133 = sqrt(m1s);
  MSf233 = sqrt(m2s);
  USf3311 = U1[0];  USf3312 = U1[1];
  USf3321 = U2[0];  USf3322 = U2[1];
  
  m11 = MSq3L*MSq3L + MB*MB + (-1/2.+1/3.*SWs)*C2B*MZs;
  m12 = MB * (Ab - mu*TB);
  m22 = MSd3R*MSd3R + MB*MB - 1/3.*SWs*C2B*MZs;
  eigenvalues(m11, m12, m12, m22, m1s, m2s);
  U1[1] = m11 - m1s;
  U1[0] = -m12;
  U2[1] = m11 - m2s;
  U2[0] = -m12;
  normalize(2, U1);
  normalize(2, U2);
  MSf143 = sqrt(m1s);
  MSf243 = sqrt(m2s);
  USf4311 = U1[0];  USf4312 = U1[1];
  USf4321 = U2[0];  USf4322 = U2[1];
  
  double mn[4][4] = {{ M1, 0, -MZ*SW*CB, MZ*SW*SB },
  		     { 0, M2,  MZ*CW*CB,-MZ*CW*SB },
		     {-MZ*SW*CB, MZ*CW*CB, 0, -mu },
		     { MZ*SW*SB,-MZ*CW*SB, -mu, 0 }};
  double defpol[5] = { mu*(-(M1*M2*mu) + 2*CB*SB*MZs*(CW*CW*M1 + M2*SWs)),
    		       M1*(mu*mu + CW*CW*MZs) - 2*CB*mu*MZs*SB + 
    			M2*(mu*mu + MZs*SWs),
    		       (M1*M2 - mu*mu - MZs),
    		       -M1-M2,
    		       1 };
  double neval[4] = { M1, M2, mu, -mu };
  double nevec[4][4], mn2[4][4];
  double ev, ev2, ev3, evold, absv;
  int n, i, j, k;
  
  cout.precision(8);
/*  for(i = 0; i < 4; i++)
  {
    ev = neval[i];  evold = 0;
    for(j = 0; j < 6 && abs((ev-evold)/ev) > 5.0e-16; j++)
    {
      ev2 = ev*ev;
      ev3 = ev2*ev;
      evold = ev;
      ev -= (defpol[0] + defpol[1]*ev + defpol[2]*ev2 + defpol[3]*ev3 +
	      defpol[4]*ev3*ev) /
	     (defpol[1] + 2*defpol[2]*ev + 3*defpol[3]*ev2 + 4*defpol[4]*ev3);
    }
    neval[i] = ev;
  }*/
  int iq,ip;
  double tresh,theta,tau,t,sm,s,h,g,c,b[4],z[4],a[4][4];
  
  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      a[i][j] = mn[i][j];
  for (ip=0;ip<4;ip++) 
  {
    b[ip]=neval[ip]=a[ip][ip];
    z[ip]=0.0;
  }
  for (i=1;i<=50;i++)
  {
    sm=0.0;
    for (ip=0;ip<3;ip++)
    {
      for (iq=ip+1;iq<3;iq++)
	sm += abs(a[ip][iq]);
    }
    if(abs(sm) < 5.0e-16)
      break;
    if (i < 4)
      tresh=0.2*sm/(4*4);
    else
      tresh=0.0;
    for (ip=0;ip<3;ip++)
    { 
      for (iq=ip+1;iq<4;iq++) 
      { 
	g=100.0*fabs(a[ip][iq]);
	if (i > 4 && (float)(fabs(neval[ip])+g) == (float)fabs(neval[ip]) 
		  && (float)(fabs(neval[iq])+g) == (float)fabs(neval[iq])) 
	{
	  a[ip][iq]=0.0;
	}
	else
	{
	  if (abs(a[ip][iq]) > tresh)
	  {
	    h=neval[iq]-neval[ip]; 
	    if ((float)(fabs(h)+g) == (float)fabs(h)) 
	      t=(a[ip][iq])/h;
	    else
	    {
	      theta=0.5*h/(a[ip][iq]); 
	      t=1.0/(abs(theta)+sqrt(1.0+theta*theta)); 
	      if (theta < 0.0) t = -t;
	    }
	    c=1.0/sqrt(1+t*t); 
	    s=t*c; 
	    tau=s/(1.0+c); 
	    h=t*a[ip][iq]; 
	    z[ip] -= h; 
	    z[iq] += h; 
	    neval[ip] -= h; 
	    neval[iq] += h; 
	    a[ip][iq]=0.0;
	    for (j=0;j<=ip-1;j++)
	      ROTATE(a,j,ip,j,iq);
	    for (j=ip+1;j<=iq-1;j++)
	      ROTATE(a,ip,j,j,iq);
	    for (j=iq+1;j<4;j++)
	      ROTATE(a,ip,j,iq,j);
	  }
	}
      }
    }
    for (ip=0;ip<4;ip++)
    { 
      b[ip] += z[ip]; 
      neval[ip]=b[ip];
      z[ip]=0.0;
    }
  }
  
  for(i = 0; i < 4; i++)
    for(j = i+1; j < 4; j++)
     if(abs(neval[i]) > abs(neval[j]))
      {
	ev = neval[i];
	neval[i] = neval[j];
	neval[j] = ev;
      }
  MNeu1 = neval[0];  MNeu2 = neval[1];  MNeu3 = neval[2];  MNeu4 = neval[3]; 

  for(n = 0; n < 4; n++)
  {
    for(i = 0; i < 4; i++)
      for(j = 0; j < 4; j++)
	mn2[i][j] = mn[i][j];
    for(i = 0; i < 4; i++)
      mn2[i][i] = mn[i][i] - neval[n];

    for(i = 0; i < 4; i++)
    {
      ev = mn2[i][i];
      for(j = 0; j < 4; j++)
	mn2[i][j] /= ev;
      for(j = i+1; j < 4; j++)
      {
	ev = mn2[j][i];
	for(k = 0; k < 4; k++)
	  mn2[j][k] -= mn2[i][k]*ev;
      }
    }
    nevec[n][3] = 1.0;
    absv = 0.0;
    for(i = 2; i >= 0; i--)
    {
      nevec[n][i] = 0.0;
      for(j = 3; j > i; j--)
	nevec[n][i] -= mn2[i][j]*nevec[n][j];
      absv += nevec[n][i]*nevec[n][i];
    }
    absv = sqrt(absv+1.0);
    for(i = 0; i < 4; i++)
      nevec[n][i] /= absv;

  }
  ZNeu11=nevec[0][0]; ZNeu12=nevec[0][1]; ZNeu13=nevec[0][2]; ZNeu14=nevec[0][3];
  ZNeu21=nevec[1][0]; ZNeu22=nevec[1][1]; ZNeu23=nevec[1][2]; ZNeu24=nevec[1][3];
  ZNeu31=nevec[2][0]; ZNeu32=nevec[2][1]; ZNeu33=nevec[2][2]; ZNeu34=nevec[2][3];
  ZNeu41=nevec[3][0]; ZNeu42=nevec[3][1]; ZNeu43=nevec[3][2]; ZNeu44=nevec[3][3];
}
